from lingvo.core import py_utils
from lingvo.core import test_utils
import numpy as np

from sidlingvo import fe_utils


class FrontendTest(test_utils.TestCase):

  def test_frontend(self):
    """Test the frontend."""
    input_signal = np.random.randn(1, 16000)
    input_paddings = fe_utils.np.zeros(input_signal.shape[:2])

    fe = fe_utils.get_frontend_layer()
    result = fe.FProp(
        fe.theta,
        py_utils.NestedMap(src_inputs=input_signal, paddings=input_paddings),
    )
    features = result.src_inputs[0, :, :, 0].numpy()
    self.assertEqual(features.shape, (33, 512))



"""Layer to perform online frame-based averaging.
"""

from lingvo import compat as tf
from lingvo.core import base_layer
from lingvo.core import layers as lingvo_layers
from lingvo.core import py_utils


class CumulativeStatisticsLayer(base_layer.BaseLayer):
  """Online layer to calculate cumulative statistics of the input over time."""

  @classmethod
  def Params(cls):
    p = super(CumulativeStatisticsLayer, cls).Params()
    p.name = 'cumulative_statistics_layer'

    p.Define(
        'stats_type', 'PASS_THRU',
        'If <PASS_THRU> then overrides all other options and copies the input '
        'directly to the output. If <MEAN> then calculates the mean statistics.'
        ' If <MEAN_STD> then calculates the mean and standard deviation '
        'statistics.')
    p.Define(
        'use_weighted_frames', False,
        'If True, use a feedforward network to weight the frames for statistics'
        ' accumulation purposes. It is recommended that a sigmoid output is '
        'used so that counts are handled correctly. If False, a simple average '
        'is used. This is only applicable for non-pass-thru configurations.')
    p.Define(
        'features_name', 'encoded',
        'The key in input_batch pointing to the features. The output features '
        'after FProp will be stored with the same name.')
    p.Define(
        'paddings_name', 'padding',
        'The key in input_batch pointing to the paddings. (Paddings are given '
        'as 0 for speech, 1 for non-speech.) The output paddings after FProp '
        'will be stored with the same name.')
    p.Define(
        'input_dim', None, 'Size of the input features to the network. Must be '
        'set to a value greater than 0. It is used for defining/retrieving the '
        'inference states shape.')
    p.Define(
        'epsilon', 1e-4,
        'This is used for 2 reasons. (1) This small value is added to the count'
        ' for each frame as part of the overall frame count statistics. It is '
        'to ensure no division by zero issues. (2) This same value is added to '
        'each variance element in the standard deviation calculation. It is to '
        'ensure that we do not compute the square root of a negative number.')
    p.Define(
        'frame_weight_ffn', lingvo_layers.FeedForwardNet.Params(),
        'The FeedForwardNet params for the frame-weighting layer. This network '
        'is activated when the use_weighted_frames flag is set to True.')

    return p

  def __init__(self, params):
    super().__init__(params)
    p = self.params

    # Check that the statistics type is valid:
    # We are checking that the statistics type is one of the three types
    # because we do not explicitly check for the 'MEAN' statistics type later
    # in the code.
    if p.stats_type not in ['PASS_THRU', 'MEAN', 'MEAN_STD']:
      raise ValueError('The stats_type must be <PASS_THRU>, '
                       '<MEAN>, or <MEAN_STD>.')

    # Check that an input dimension is specified
    if p.input_dim is None and p.stats_type != 'PASS_THRU':
      raise ValueError('The input dimension (input_dim) must be explicitly set '
                       'and specifies the dimensionality of a frame of input '
                       'features. The value must also be greater than 0 (and '
                       'not None).')

    # Create the feedforward layers for the frame-weighting result
    if p.use_weighted_frames:
      self.CreateChild('frame_weight_ffn', p.frame_weight_ffn)

  def zero_state(self, theta, batch_size):
    """Returns the initial state of the current layer.

    Args:
      theta: A NestedMap containing layer weights (unused for this function).
      batch_size: The number of examples (utterances) in the minibatch

    Returns:
      A NestedMap containing the initial state of the sufficient statistics. If
      the statistics type (stats_type) is set to 'PASS_THRU', then an empty
      NestedMap is returned. If the statistics type is set to 'MEAN', then
      the count and sum of X information is returned. If the statistics type is
      set to 'MEAN_STD', then the sum of X squared statistics are included in
      addition to the count and sum of X information. Depending on configuration
      settings, the NestedMap can contain the following keys within the
      NestedMap accumulated_stats:
      - count: tf.zeros([batch_size]), fprop_dtype (default tf.float32)
      - sum_x: tf.zeros([batch_size, input_dim]), fprop_dtype (default
      tf.float32)
      - sum_xx: tf.zeros([batch_size, input_dim]), fprop_dtype (default
      tf.float32)
    """

    p = self.params

    state0 = py_utils.NestedMap()

    # Return the initial state for PASS_THRU mode
    if p.stats_type == 'PASS_THRU':
      return state0

    # Specify the initial state
    state0.accumulated_stats = py_utils.NestedMap()
    state0.accumulated_stats.count = tf.zeros([batch_size], dtype=p.dtype)
    state0.accumulated_stats.sum_x = tf.zeros([batch_size, p.input_dim],
                                              dtype=p.dtype)

    # Add zero state for sum x-squared statistics if requested
    if p.stats_type == 'MEAN_STD':
      state0.accumulated_stats.sum_xx = tf.zeros([batch_size, p.input_dim],
                                                 dtype=p.dtype)

    return state0

  def IsNullState(self, state):
    """Checks if the input state is either None or empty NestedMap."""
    return all(x is None for x in tf.nest.flatten(state))

  def NullState(self):
    """Returns empty NestedMap as null state."""
    return py_utils.NestedMap()

  def FProp(self, theta, in_nmap, state0=None):
    """Generates frame-weighted mean/std-dev statistics from the inputs.

    Args:
      theta: A NestedMap containing layer weights containing the key
        frame_weight_ffn describing the feed-forward network. This key is needed
        only when use_weighted_frames is set to True and stats_type is not
        'PASS_THRU'.
      in_nmap: A NestedMap. Members include:
        - in_map[p.features_name]: Features tensor of shape [len, batch,
          input_dim], tf.float32.
        - in_map[p.paddings_name]: Paddings tensor of shape [len, batch],
          tf.float32.
      state0: A NestedMap containing sufficient statistics for the previous
        state. When not in inference mode state0 should be NullState. When in
        inference mode, state0, in addition to containing state0 information
        from other child layers, should also include the following keys within
        the NestedMap state0.accumulated_stats:
        - count: [batch_size], tf.float32
        - sum_x: [batch_size, input_dim], tf.float32
        - sum_xx: [batch_size, input_dim], tf.float32 The above keys point to
          the sufficient statistics accumulated across all data packets
          excluding the current data packet of features.

    Returns:
      A NestedMap (out_nmap). For the 'PASS_THRU' case, in_nmap is returned. For
      the 'MEAN' and 'MEAN_STD' cases, a NestedMap with the same information and
        structure as in_nmap with the following additional updates:
        out_nmap[p.features_name]: Features tensor of shape [len, batch,
          output_dim], tf.float32. The output_dim is either input_dim or
          2*input_dim depending on whether standard deviation statistics are
          also included.
        out_nmap[p.paddings_name]: Paddings tensor of shape [len, batch],
          tf.float32.
        out_nmap.state: If state0 is NullState, NullState is returned as the
          output state. If state0 is not NullState, then a NestedMap containing
          sufficient statistics of this cumulative statistics layer gets
          returned. When the mode is 'PASS_THRU', the state is an empty
          NestedMap. In other cases, the state includes the frame-based counts,
          sum of X and sum of X-squared (if 'MEAN_STD' information is requested)
          The additional NestedMap keys are included within
          out_nmap.state.accumulated_stats as follows:
          - count: [batch_size], tf.float32
          - sum_x: [batch_size, input_dim], tf.float32
          - sum_xx: [batch_size, input_dim], tf.float32
          For example, to reference the counts, given an output variable,
          out_nmap, the following would be used:
          out_nmap.state.accumulated_stats.count
    """

    p = self.params

    # Return the input if we are using PASS_THRU mode.
    # Note: the state, even though it is empty, is always populated so as to be
    # consistent with the non-PASS_THRU cases.
    if p.stats_type == 'PASS_THRU':
      # Do not mutate the input NestedMap but the state, because of PASS_THRU.
      out_nmap = in_nmap.copy()
      out_nmap.state = self.NullState()
      return out_nmap

    # Get the input data and padding information
    input_features = in_nmap[p.features_name]
    padding = in_nmap[p.paddings_name]

    # Convert padding to frame based flag indicating if it is speech
    effective_frame_weight = tf.cast(1.0 - padding, dtype=p.dtype)

    # If using frame weighted analysis, calculate the sigmoid output and
    # multiply it with the speech/non-speech effective_frame_weight.
    if p.use_weighted_frames:
      # For each speech frame, a single weight value is generated between 0 and
      # 1 (if a sigmoid activation is used). The expected output (after the
      # squeeze function) is a tensor of shape [len, batch] with a tf.float32
      # data type.
      ffn_frame_weight = tf.squeeze(
          self.frame_weight_ffn.FProp(theta.frame_weight_ffn, input_features,
                                      None),
          axis=2)

      effective_frame_weight = effective_frame_weight * ffn_frame_weight

    # Add a small floor for the effective_frame_weight
    effective_frame_weight = effective_frame_weight + p.epsilon

    # Calculate the cumulative count and cumulative sum_x for the current packet
    # of frames
    cumulative_count = tf.math.cumsum(effective_frame_weight, axis=0)
    cumulative_sum_x = tf.math.cumsum(
        input_features * effective_frame_weight[:, :, tf.newaxis], axis=0)

    # If standard deviation statistics are needed, calculate the cumulative
    # sum_xx (sum x-squared) for the current packet of frames
    if p.stats_type == 'MEAN_STD':
      cumulative_sum_xx = tf.math.cumsum(
          input_features * input_features *
          effective_frame_weight[:, :, tf.newaxis],
          axis=0)

    state1 = self.NullState()
    # If we are running in online mode, be sure to add in the total sums from
    # the past packets of features. If in offline mode, there is no state to
    # update.
    if not self.IsNullState(state0):
      # Calculate cumulative sums up to the current point in time. This includes
      # past packets and the current packet of data.
      cumulative_count = cumulative_count + state0.accumulated_stats.count[
          tf.newaxis, :]
      cumulative_sum_x = cumulative_sum_x + state0.accumulated_stats.sum_x[
          tf.newaxis, :, :]

      # Update the internal state by copying across the statistics of the very
      # last frame
      state1.accumulated_stats = py_utils.NestedMap()
      state1.accumulated_stats.count = cumulative_count[-1, :]
      state1.accumulated_stats.sum_x = cumulative_sum_x[-1, :, :]

      # If standard deviation statistics are needed, calculate the cumulative
      # sum_xx (sum x-squared) for the past and the current packet of frames
      if p.stats_type == 'MEAN_STD':
        cumulative_sum_xx = cumulative_sum_xx + state0.accumulated_stats.sum_xx[
            tf.newaxis, :, :]
        state1.accumulated_stats.sum_xx = cumulative_sum_xx[-1, :, :]

    # Calculate the running mean for the current packet of features
    output_features = tf.math.divide(cumulative_sum_x,
                                     cumulative_count[:, :, tf.newaxis])

    # If requested, calculate and append the standard deviation statistics.
    if p.stats_type == 'MEAN_STD':
      cumulative_mean_xx = tf.math.divide(cumulative_sum_xx,
                                          cumulative_count[:, :, tf.newaxis])
      cumulative_std_dev = tf.math.sqrt(cumulative_mean_xx -
                                        output_features * output_features +
                                        p.epsilon)
      output_features = tf.concat((output_features, cumulative_std_dev), axis=2)

    # Return the output
    out_nmap = py_utils.NestedMap()
    out_nmap[p.features_name] = output_features
    out_nmap[p.paddings_name] = padding
    out_nmap.state = state1
    return out_nmap



"""__init__ file."""

from . import fe_utils
from . import wav_to_dvector
from . import wav_to_lang

load_tflite_model = fe_utils.load_tflite_model

WavToLangRunner = wav_to_lang.WavToLangRunner

WavToDvectorRunner = wav_to_dvector.WavToDvectorRunner



"""Tests for attentive scoring."""

from sidlingvo import attentive_scoring_layer
from lingvo import compat as tf
from lingvo.core import test_utils
import numpy as np

_TF_RANDOM_SEED = 1314
_ABS_TOLERANCE = 1e-05
_REL_TOLERANCE = 1e-04


class AttentiveScoringLayerTest(test_utils.TestCase):
  """The attentive scoring layer tests."""

  def setUp(self):
    """Generates test examples and the attentive scoring configuration."""

    super().setUp()

    tf.reset_default_graph()

    # Generate the example to test the attentive scoring
    (self.test_data, self.enroll_data, self.data_info) = self._GenerateExample()

    # Set up the basic attentive scoring layer config
    p = attentive_scoring_layer.AttentiveScoringLayer.Params()
    p.num_keys = self.data_info['num_keys']
    p.key_dim = self.data_info['key_dim']
    p.value_dim = self.data_info['value_dim']
    p.scale_factor = 1.0
    p.use_trainable_scale_factor = False
    p.apply_l2_norm_to_keys = False
    p.apply_l2_norm_to_values = True
    p.apply_global_l2_norm_to_concat_form = False

    self.params = p

  def _GenerateExample(self):
    """Generates an example to use in the tests.

    This example has 2 enrollment speakers (2 utterances per speaker), and 2
    test utterances. There are 2 keys used. Keys are 2 dimensional and values
    are 3 dimensional.

    Given these constraints, utterance representations are given as:
    Feature indexes | Description
             0 to 1 | First key vector (length 2)
             2 to 4 | First value vector (length 3)
             5 to 6 | Second key vector (length 2)
             7 to 9 | Second value vector (length 3)

    Returns:
      A tuple of (test_data, enroll_data, data_info)
      test_data: Test data related to 2 test utterances with 2 key vectors of 2
        dimensions and 2 value vectors of 3 dimensions. Each utterance
        representation contains a packed form of the key and value vectors. The
        result is a 2d tensor (of tf.float32 type elements) of dimension
        [num_test_utts, representation_dim]. In this example, num_test_utts
        is 2 and representation_dim is 10 (or 2 keys * (2 key_dim +
        3 value_dim)).
      enroll_data: Enrollment data related to 2 speakers each with 2 enrollment
        utterances. Each utterance is composed of 2 key vectors of 2 dimensions
        and 2 value vectors of 3 dimensions. Each utterance is a packed form of
        the key and value vectors. The result is a 3d tensor (of tf.float32
        type elements) of dimension [num_enroll_spks, num_enroll_utts_per_spk,
        representation_dim].
      data_info: A dictionary with keys containing the relevant information for
        the returned data. The dictionary keys include num_keys, key_dim,
        value_dim, num_enroll_spks, num_enroll_utts_per_spk, representation_dim
        and num_test_utts. All values are Python integers.
    """

    # Generate test utterance representations in packed key/value form.
    test_utt1 = np.array([1, 2, -1, 2, 6, 3, -4, 5, 10, -2], dtype=np.float32)
    test_utt2 = np.array([3, -5, -10, 4, 7, 2, 7, 3, -12, -3], dtype=np.float32)

    # Form the test data. They are 2d tensors of tf.float32 and shape:
    # ([num_test_utts, representation_dim])
    test_data = tf.convert_to_tensor(np.stack((test_utt1, test_utt2), axis=0))

    # Generate enrollment utterance representations in packed key/value form.
    # Each utterance is packed in the same way the test utterances are.
    enroll_spk1_utt1 = np.array([1, 2, -1, 2, 6, 3, -4, 5, 10, -2],
                                dtype=np.float32)
    enroll_spk1_utt2 = np.array([3, 5, -2, 3, -6, -1, -5, -5, 9, -7],
                                dtype=np.float32)
    enroll_spk2_utt1 = np.array([5, -2, -3, 7, 5, -6, 8, 1, 3, -7],
                                dtype=np.float32)
    enroll_spk2_utt2 = np.array([4, 5, -10, 8, 7, 3, 2, 7, 2, -5],
                                dtype=np.float32)

    # Form the enrollment data. They are 3d tensors of tf.float32 and shape:
    # [num_enroll_spks, num_enroll_utts_per_spk, representation_dim]
    enroll_data = tf.convert_to_tensor(
        np.stack((np.stack((enroll_spk1_utt1, enroll_spk1_utt2), axis=0),
                  np.stack((enroll_spk2_utt1, enroll_spk2_utt2), axis=0)),
                 axis=0))

    # Specify the data configuration
    data_info = {}
    data_info['num_keys'] = 2
    data_info['key_dim'] = 2
    data_info['value_dim'] = 3
    data_info['num_enroll_spks'] = 2
    data_info['num_enroll_utts_per_spk'] = 2
    data_info['representation_dim'] = 10
    data_info['num_test_utts'] = 2

    return (test_data, enroll_data, data_info)

  def _TestHelper(self, params, test_data, enroll_data):
    """Returns the attentive scores for the given test and enrollment data.

    Args:
      params: Babelfish configuration parameters for setting up the
        attentive_scoring_layer.
      test_data: Test data related to 2 test utterances each with 2 key vectors
        of 2 dimensions and 2 value vectors of 3 dimensions. Each utterance
        representation contains a packed form of the key and value vectors. The
        result is a 2d tensor (of tf.float32 elements) of dimension
        [num_test_utts, representation_dim]. In this example, num_test_utts is 2
        and representation_dim is 10 (or 2 keys * (2 key_dim + 3 value_dim)).
      enroll_data: Enrollment data related to 2 speakers each with 2 enrollment
        utterances. Each utterance is composed of 2 key vectors of 2 dimensions
        and 2 value vectors of 3 dimensions. Each utterance is a packed form of
        the key and value vectors. The result is a 3d tensor (of tf.float32
        elements) of dimension [num_enroll_spks, num_enroll_utts_per_spk,
        representation_dim].

    Returns:
      The output of the attentive scoring. The result is a numpy np.float32
      tensor of shape [num_test_utts, num_enroll_spks].
    """

    with self.session() as sess:
      tf.random.set_seed(_TF_RANDOM_SEED)
      attention_network = params.Instantiate()

      output = attention_network.FProp((test_data, enroll_data))

      sess.run(
          tf.group(tf.global_variables_initializer(), tf.tables_initializer()))

      return sess.run(output)

  def testAttentionNetworkFProp(self):
    """Checks that the forward propagation is correct for attentive modeling."""

    # Run the test
    output = self._TestHelper(self.params, self.test_data, self.enroll_data)

    expected_output = np.array([[0.99984203, 0.43492252],
                                [-0.26937336, -0.30256001]])
    self.assertAllClose(
        expected_output, output, rtol=_REL_TOLERANCE, atol=_ABS_TOLERANCE)

  def testAttentionNetworkFPropWithPerTestKeySoftmax(self):
    """Checks forward propagation result for softmax calculated per test key."""

    # Calculate the softmax for each test utterance key rather than a softmax
    # across all enrollment and test keys for a trial.
    self.params.apply_softmax_per_test_key = True

    # Run the test
    output = self._TestHelper(self.params, self.test_data, self.enroll_data)

    expected_output = np.array([[0.18785089, 0.5673896],
                                [-0.21675828, 0.24606878]])
    self.assertAllClose(
        expected_output, output, rtol=_REL_TOLERANCE, atol=_ABS_TOLERANCE)

  def testAttentionNetworkFPropTrainableScaleFactor(self):
    """Checks that the forward propagation is correct for trainable scaling."""

    # Enable trainable scaling
    self.params.use_trainable_scale_factor = True
    self.params.scale_factor = 2.0

    with self.session() as sess:
      tf.random.set_seed(_TF_RANDOM_SEED)
      attention_network = self.params.Instantiate()

      output = attention_network.FProp((self.test_data, self.enroll_data),
                                       attention_network.theta)
      sess.run(
          tf.group(tf.global_variables_initializer(), tf.tables_initializer()))
      output_result = sess.run(output)
      log_scale_factor_result = sess.run(
          attention_network.theta.trainable_log_scale_factor)

    expected_output = np.array([[1.0, 0.434889], [-0.269374, -0.202443]])
    self.assertAllClose(
        expected_output,
        output_result,
        rtol=_REL_TOLERANCE,
        atol=_ABS_TOLERANCE)

    # Check that the log of the scale_factor=2 is as expected
    self.assertAllClose(
        0.6931471824645996,
        log_scale_factor_result,
        rtol=_REL_TOLERANCE,
        atol=_ABS_TOLERANCE)

  def testAttentionNetworkFPropForKeysAndQueries(self):
    """Checks forward propagation result for key and query calculations."""

    # Calculate the scores using separate keys and queries.
    self.params.use_keys_and_queries = True
    self.params.key_dim = 1

    # Run the test
    output = self._TestHelper(self.params, self.test_data, self.enroll_data)

    expected_output = np.array([[-0.227293, 0.561662], [-0.477102, -0.92821]])
    self.assertAllClose(
        expected_output, output, rtol=_REL_TOLERANCE, atol=_ABS_TOLERANCE)

  def testAttentionNetworkFPropL2NormConcatForm(self):
    """Check forward prop. for L2-norm concatenated representation form."""

    # Calculate the scores given L2-norm on the concatenated representation.
    self.params.apply_l2_norm_to_keys = False
    self.params.apply_l2_norm_to_values = False
    self.params.apply_global_l2_norm_to_concat_form = True

    # Run the test
    output = self._TestHelper(self.params, self.test_data, self.enroll_data)

    expected_output = np.array([[0.9998327494, 0.4348915219],
                                [-0.2693726420, -0.3787475228]])
    self.assertAllClose(
        expected_output, output, rtol=_REL_TOLERANCE, atol=_ABS_TOLERANCE)


if __name__ == '__main__':
  tf.test.main()



"""A library to run speaker-id TFLite model inference."""

import dataclasses
from lingvo.core import py_utils
import numpy as np
import fe_utils


def aggregate_dvectors(dvectors: list[np.ndarray]) -> np.ndarray:
  """Aggregate dvectors from multiple utterances."""
  normalized_dvectors = [
      dvector / np.linalg.norm(dvector) for dvector in dvectors
  ]
  stacked_dvectors = np.stack(normalized_dvectors, axis=0)
  return np.mean(stacked_dvectors, axis=0, keepdims=False)


def compute_cosine(vec1: np.ndarray, vec2: np.ndarray) -> float:
  """Compute cosine similarity between two vectors."""
  return np.inner(vec1, vec2) / (np.linalg.norm(vec1) * np.linalg.norm(vec2))


@dataclasses.dataclass
class WavToDvectorRunner:
  """WavToDvectorRunner."""

  # Path to the VAD TFLite model.
  vad_model_file: str = ""
  # Path to the VAD mean and stddev CSV file.
  vad_mean_stddev_file: str = ""
  # Path to the speaker-id TFLite model.
  tisid_model_file: str = ""

  vad_threshold: float = 0.1
  vad_cluster_id: int = 2
  vad_num_clusters: int = 16

  def __post_init__(self):
    self.vad_model = fe_utils.load_tflite_model(self.vad_model_file)
    self.tisid_model = fe_utils.load_tflite_model(self.tisid_model_file)
    self.vad_mean_stddev = fe_utils.read_mean_stddev_csv(
        self.vad_mean_stddev_file
    )

  def wav_to_dvector(self, audio_file: str) -> np.ndarray:
    """Run speaker-id model on wav file."""
    input_signal = fe_utils.get_int_samples(audio_file)
    return self.samples_to_dvector(input_signal)

  def samples_to_dvector(self, input_signal: np.ndarray) -> np.ndarray:
    """Run speaker-id model on int16 samples."""
    input_paddings = fe_utils.np.zeros(input_signal.shape[:2])
    fe = fe_utils.get_frontend_layer()
    result = fe.FProp(
        fe.theta,
        py_utils.NestedMap(src_inputs=input_signal, paddings=input_paddings),
    )
    features = result.src_inputs[0, :, :, 0].numpy()
    self.vad_model.reset_all_variables()
    _, features_after_vad = fe_utils.apply_vad(
        features,
        self.vad_model,
        self.vad_mean_stddev,
        self.vad_threshold,
        self.vad_cluster_id,
        self.vad_num_clusters,
    )
    self.tisid_model.reset_all_variables()
    dvectors = fe_utils.run_multi_input_model(
        self.tisid_model, features_after_vad
    )
    return dvectors

  def compute_score(
      self, enroll_audio_list: list[str], test_audio: str
  ) -> float:
    """Compute the cosine similarity score between enroll and test audio."""
    enroll_dvectors = [
        self.wav_to_dvector(enroll_audio)[-1, :]
        for enroll_audio in enroll_audio_list
    ]
    aggregate_enroll_dvector = aggregate_dvectors(enroll_dvectors)
    test_dvector = self.wav_to_dvector(test_audio)[-1, :]
    score = compute_cosine(aggregate_enroll_dvector, test_dvector)
    return score



"""Speaker recognition loss layers."""

import enum
from lingvo import compat as tf
from lingvo.core import base_layer
from lingvo.core import layers as lingvo_layers
from lingvo.core import py_utils
import attentive_scoring_layer
from sidlingvo import utils


class EmbeddingComparisonType(enum.Enum):
  """Add enum capabilities for selecting cosine or attentive scoring."""

  # Instructs the system to perform cosine scoring of the test utterance
  # embedding against the averaged speaker enrollment utterance embedding.
  COSINE = enum.auto()

  # Performs attentive scoring of the test utterance against all enrollment
  # utterances. Attentive scoring involves the use of keys/queries and values to
  # compare a test utterance with multiple enrollment utterances. For more
  # information, refer to this paper:
  # https://arxiv.org/abs/2203.05642
  ATTENTIVE_SCORING = enum.auto()


class GEnd2EndSoftmaxLayer(base_layer.BaseLayer):
  """The generalized end-to-end softmax loss layer.

  This is implemented based on Eq.(6) proposed in the GEnd2End paper:
  https://arxiv.org/pdf/1710.10467.pdf
  """

  @classmethod
  def Params(cls):
    p = super(GEnd2EndSoftmaxLayer, cls).Params()
    # Below values should match the input generator params!
    p.Define('batch_size', 128, 'The batch size.')
    p.Define('num_spks_per_batch', 16, 'The number of speakers in the batch.')
    p.Define('num_utts_per_spk', 8, 'The number of utts per speaker.')
    p.Define('vary_number_of_enrollment_utterances_averaged', False,
             ('Vary the number of enrollment utterances in training (True) or '
              'use all the available enrollment utterances (default: False).'))
    p.Define('softmax', lingvo_layers.SimpleFullSoftmax.Params(),
             'Softmax layer.')
    p.Define(
        'split_batch', True, 'Whether to split batch into enrollment and '
        'test at loss computation.')
    p.Define(
        'select_embedding_comparison_type', EmbeddingComparisonType.COSINE,
        'Determines what method should be used to compare the enrollment '
        'and test utterance embeddings. There are currently 2 options: '
        'EmbeddingComparisonType.COSINE and '
        'EmbeddingComparisonType.ATTENTIVE_SCORING. (Default: '
        'EmbeddingComparisonType.COSINE)')
    p.Define('attentive_scoring',
             attentive_scoring_layer.AttentiveScoringLayer.Params(),
             'Attentive scoring layer.')
    p.Define('transform_weight', 10.0, 'The transform weight initial value.')
    p.Define('transform_bias', -5.0, 'The transform bias initial value.')
    p.Define('mask_dup_spk_scores', False,
             'Whether to mask duplicate speaker scores in the batch.')
    p.Define(
        'mask_value', 100.0,
        'The masking value to be subtracted from the scores matrix.'
        'Note that this is done after the linear transformation of the'
        'similarity scores.')
    p.name = 'gend2end_softmax_layer'
    return p

  def __init__(self, params):
    if params.num_utts_per_spk % 2 != 0:
      raise ValueError('An even number of examples per speaker is expected, '
                       'but got %i.)' % (params.num_utts_per_spk))

    if params.batch_size != params.num_spks_per_batch * params.num_utts_per_spk:
      raise ValueError(
          'The batch size must be the same as the number of '
          'speakers times the number of examples per speaker. But '
          'batch_size != num_spks_per_batch * num_utts_per_spk, %i != %i * %i)'
          % (params.batch_size, params.num_spks_per_batch,
             params.num_utts_per_spk))

    super(GEnd2EndSoftmaxLayer, self).__init__(params)
    p = self.params
    self.CreateChild('softmax', p.softmax)
    self.CreateChild('attentive_scoring', p.attentive_scoring)

  def _CreateLayerVariables(self):
    super()._CreateLayerVariables()
    p = self.params
    pc = py_utils.WeightParams(
        shape=[],
        init=py_utils.WeightInit.Constant(scale=p.transform_weight),
        dtype=p.dtype,
        collections=[self.__class__.__name__ + '_vars'])
    self.CreateVariable('w', pc)
    pc = py_utils.WeightParams(
        shape=[],
        init=py_utils.WeightInit.Constant(scale=p.transform_bias),
        dtype=p.dtype,
        collections=[self.__class__.__name__ + '_vars'])
    self.CreateVariable('b', pc)

  def BatchWithoutDuplicateSpeakers(self, input_batch):
    """Check if the batch has duplicate speakers.

    This tensor is multiplied on top of the default loss weight of batch size.
    If the input batch contains duplicate speakers, we return 0, so that the
    batch loss gets weighted 0. Otherwise, the batch is valid and we return 1.

    Args:
      input_batch: A NestedMap containing:
        - src.src_inputs: tensor of shape [batch, len, dim].
        - src.paddings: tensor of shape [batch, len].
        - label: int tensor of shape [batch]. The inputs has ground truth labels
          required for the loss calculation.

    Returns:
      A int32 tensor of value 0 for invalid batches with duplicate speakers,
      and 1 for valid ones without duplicate speakers.
    """
    p = self.params

    # Collect all the speakers
    speakers = tf.gather(
        tf.reshape(input_batch.label,
                   [p.num_spks_per_batch, p.num_utts_per_spk]), [0],
        axis=1)
    speakers = tf.squeeze(speakers)
    # Check the batch speaker compositions
    return tf.cast(
        tf.reduce_all(tf.equal(speakers, tf.range(p.num_spks_per_batch))),
        dtype=tf.int32)

  def MaskDuplicateSpeakerScores(self, score, input_batch):
    """Mask duplicate speaker scores in the score matrix and update the labels.

    Illustration on a 6*3 matrix of 3 speakers.
      1 [(11)  12   13*]
      2 [(21)  22   23*]
      3 [ 31  (32)  33 ]
      4 [ 41  (42)  43 ]
      5 [ 51*  52  (53)]
      6 [ 61*  62  (63)]
    The label tensor in this example is [0, 0, 1, 1, 0, 0] where the 5th and
    6th examples are from the same speaker of the 1st and 2nd examples. We mask
    the non-diagnoal, duplicate speaker target trials to be non-target, as
    shown in the matrix where *'s are. The brackets () represents for the
    target trials that remain untouched and those would be the only target
    trials that are considered in the following loss computations.

    The masking is subtracting a relatively large numeric value from the scores.

    Args:
      score: tensor of shape [batch, num_spks_per_batch] with scores of each
        example to each speaker.
      input_batch: A NestedMap containing:
        - src.src_inputs: tensor of shape [batch, len, dim].
        - src.paddings: tensor of shape [batch, len].
        - label: int tensor of shape [batch]. The inputs has ground truth labels
          required for the loss calculation.

    Returns:
      A tuple of (masked_score tensor, updated input_batch NestedMap), where
      the scores are masked and the input_batch's label tensor is also updated
      to be purely speaker indices (because the duplicate target trials have
      been masked and are now non-target trials).
    """
    p = self.params

    # Ensure the input shapes are correct
    assert len(py_utils.GetShape(score)) == 2
    assert len(py_utils.GetShape(input_batch.label)) == 1
    assert py_utils.GetShape(score)[0] == py_utils.GetShape(
        input_batch.label)[0]

    # Collect speakers
    label_2d = tf.reshape(input_batch.label,
                          [p.num_spks_per_batch, p.num_utts_per_spk])
    speakers = tf.gather(label_2d, [0], axis=1)
    speakers = tf.squeeze(speakers)

    # Create matrix for all target trials including duplicate speaker trials
    all_target_trials = tf.equal(
        tf.expand_dims(speakers, 0), tf.expand_dims(speakers, 1))

    # Create matrix for only the same speaker target trials, the diagonal
    idx = tf.cumsum(tf.ones_like(speakers)) - 1
    same_speaker_target_trials = tf.equal(
        tf.expand_dims(idx, 0), tf.expand_dims(idx, 1))

    # Calculate the maskings for the duplicate speaker target trials, which
    # we mask to treat as non-target trials.
    maskings = tf.cast(all_target_trials, tf.float32) - tf.cast(
        same_speaker_target_trials, tf.float32)

    # Repeat each row `num_utts_per_spk` times.
    maskings = tf.tile(maskings, tf.constant([1, p.num_utts_per_spk]))
    maskings = tf.reshape(
        maskings,
        [p.num_spks_per_batch * p.num_utts_per_spk, p.num_spks_per_batch])
    maskings *= p.mask_value

    # Mask the scores
    score -= maskings

    # Update the label tensor accordingly as well.
    # Note that this updated label tensor might or might not be used
    # depending on the use case. For example, for the non-MR GE2E loss, we
    # have to use this for the cross-entropy loss, but for the MR loss, this
    # is not used at all, because MR has its own transformation.
    spk_idx = tf.range(p.num_spks_per_batch)
    spk_idx = tf.reshape(spk_idx, [-1, 1])
    spk_idx = tf.tile(spk_idx, [1, p.num_utts_per_spk])
    spk_idx = tf.reshape(spk_idx, [p.num_spks_per_batch * p.num_utts_per_spk])
    assert py_utils.GetShape(input_batch.label) == py_utils.GetShape(spk_idx)
    input_batch.label = spk_idx

    # Verify that after masking there must be no duplicates.
    input_batch.label = py_utils.with_dependencies([
        py_utils.assert_equal(
            self.BatchWithoutDuplicateSpeakers(input_batch), 1)
    ], input_batch.label)

    return score, input_batch

  def ComputeSimilarity(self, logits, theta=None):
    """Top-level compute similarity score function.

    Args:
      logits: output data of network Tensor([batch, dims])
      theta: A NestedMap containing layer weights.
    Returns:
      cosine similarity matrix between each data and each speaker centroid.
      Tensor([batch, num_spks_per_batch])
    """
    p = self.params

    # [batch, num_spks_per_batch]
    if p.select_embedding_comparison_type == EmbeddingComparisonType.COSINE:
      if p.split_batch:
        score = self.ComputeSimilaritySplit(logits)
      else:
        score = self.ComputeSimilaritySimple(logits)
    elif (p.select_embedding_comparison_type ==
          EmbeddingComparisonType.ATTENTIVE_SCORING):
      score = self.ComputeAttentionScoringSimilaritySplit(logits, theta)
    else:
      raise ValueError('Invalid value for select_embedding_comparison_type.')

    return score

  def ComputeSimilaritySimple(self, logits):
    """Compute similarity score of each data to each speaker centroid.

    Each centroid is computed using all embeddings of a given speaker.

    Args:
      logits: output data of network Tensor([batch, dims])

    Returns:
      cosine similarity matrix between each data and each speaker centroid.
      Tensor([batch, num_spks_per_batch])
    """
    p = self.params
    return utils.ComputeSimilaritySimple(logits, p.num_spks_per_batch,
                                         p.num_utts_per_spk)

  def ComputeSimilaritySplit(self, logits):
    """Compute score of each data to each speaker centroid, with split batches.

    Each centroid is computed using all embeddings of a given speaker. The batch
    is split into the enrollment and test parts so that the loss is computed in
    a non-biased way.

    Args:
      logits: output data of network Tensor([batch, dims])

    Returns:
      cosine similarity matrix between each data and each speaker centroid.
      Tensor([batch, num_spks_per_batch])
    """
    p = self.params
    return utils.ComputeSimilaritySplit(
        logits, p.num_spks_per_batch, p.num_utts_per_spk,
        p.vary_number_of_enrollment_utterances_averaged)

  def ComputeAttentionScoringSimilaritySplit(self,
                                             segment_representations,
                                             theta=None):
    """Compute attention based scoring of minibatch.

    Compute the attention based score of each test segment embedding against
    each enrollment speaker (with multiple utterance embeddings), with split
    batches.

    Args:
      segment_representations: A batch of utterance representations. Tensor is
        of shape ([batch, dims]) with elements of tf.float32.
      theta: A NestedMap containing layer weights.

    Returns:
      attention based scoring similarity matrix between each test segment and
      each enrollment speaker (each enrollment speaker with their group of
      enrollment segments). Tensor([batch, num_spks_per_batch]), tf.float32.
    """
    p = self.params

    segment_representations_3d = tf.reshape(
        segment_representations, [p.num_spks_per_batch, p.num_utts_per_spk, -1])

    def _AttentionScoringGivenDataPart(part):
      """Compute attentive scoring for one of two parts.

      This helper function extracts the speaker test and enrollment
      representations according to the appropriate part selected. When part is
      0, all even numbered representations (based on index) are used as test
      segments while all odd numbered representations are used as enrollment
      speech segments. When part is 1, the assignments are switched.

      Args:
        part: Select which data part to use as test (0 or 1).

      Returns:
        scores: For a data part, calculates the attention based scores of all
          test utterances against all enrollment speakers.
          Size: [num_test_utterances, num_enroll_spks] as tf.float32
      """
      if part not in {0, 1}:
        raise ValueError('Must select an appropriate data part, either 0 or 1.')

      # Split enroll and test segments
      test_utts_3d = segment_representations_3d[:, part::2, :]
      test_utts = tf.reshape(test_utts_3d, [-1, test_utts_3d.shape[2]])

      other_part = 0 if part else 1
      enroll_utts_3d = segment_representations_3d[:, other_part::2, :]

      # Compare the test utterances against the enrollment speakers (comprising
      # multiple utterances per enrollment speaker).
      scores = self.attentive_scoring.FProp(
          inputs=(test_utts, enroll_utts_3d), theta=theta.attentive_scoring)

      return scores

    # Calculate the scores of the first set of test segments against the second
    # set of enrollment segments. Then switch them.
    scores_fst = _AttentionScoringGivenDataPart(0)
    scores_snd = _AttentionScoringGivenDataPart(1)

    # attention_similarity:
    # [batch // 2, 2, num_spks_per_batch] -> [batch, num_spks_per_batch]
    attention_similarity = tf.reshape(
        tf.stack([scores_fst, scores_snd], axis=1),
        [p.batch_size, p.num_spks_per_batch])
    return attention_similarity

  def TransformScoreAndLabel(self, theta, score, input_batch):
    """Transform the score and label tensors.

    This transformation includes the default linear transformation and masking
    if specified. This could be overridden by subclasses if there is some
    additional transformation logic, such as extended-set softmax, etc.

    Args:
      theta: A NestedMap containing layer weights.
      score: tensor of shape [batch, num_spks_per_batch] with scores of each
        example to each speaker.
      input_batch: A NestedMap containing at least:
        - label: int tensor of shape [batch]. The inputs has ground truth labels
          required for the loss calculation.

    Returns:
      A tuple of two elements:
        - A transformed score tensor, which could be of a different shape.
        - A NestedMap input_batch containing the transformed label tensor, which
          could be of a different shape.
    """
    p = self.params

    # [batch, num_spks_per_batch], linear transformation
    score = theta.w * score + theta.b

    # Mask the duplicate speaker related scores with negative values and
    # update the label tensor accordingly after masking the target trials
    # manually to be non-target trials.
    if p.mask_dup_spk_scores:
      score, input_batch = self.MaskDuplicateSpeakerScores(score, input_batch)

    return score, input_batch

  def CalculateLoss(self, theta, score, input_batch):
    """Calculate the loss given the scores and input batch labels.

    Args:
      theta: A NestedMap containing layer weights.
      score: tensor of shape [batch, num_proxy_classes] with scores of each
        example to each proxy centroid.
      input_batch: A NestedMap containing at least:
        - label: int tensor of shape [batch]. The inputs has ground truth labels
          required for the loss calculation.

    Returns:
      A tuple of two elements:
        - A transformed score tensor, which could be of a different shape.
        - A NestedMap input_batch containing the transformed label tensor, which
          could be of a different shape.
    """
    p = self.params

    # [batch]
    per_example_xent, _ = self.softmax.XentLossFromLogits(
        theta=theta.softmax,
        logits=score,
        class_weights=tf.ones_like(score.shape[1], dtype=tf.float32),
        class_ids=input_batch.label)
    avg_xent = tf.reduce_mean(per_example_xent, name='batch_loss_mean')

    metrics = py_utils.NestedMap()
    batch_size = py_utils.GetShape(input_batch.label)[0]
    # Convert the weight of batch_size into a tensor to prevent test time error
    weight = tf.constant(batch_size)
    # When we skip masking duplicate speaker scores, we need to assign zero
    # weight to the batches with duplicate speakers.
    if not p.mask_dup_spk_scores:
      weight *= self.batch_without_duplicate_speakers
    metrics.loss = (avg_xent, weight)
    return metrics

  def FProp(self, theta, predictions, input_batch):
    """Computes the GEnd2End softmax loss and other metrics.

    Args:
      theta: A NestedMap containing layer weights.
      predictions: A NestedMaps of the predictions:
        - logits: tensor of shape [batch, dvector_size].
        - paddings: tensor of shape [batch, len].
        - state1: RNN output states, a NestedMap, for each RNN layer containing:
          - m: the lstm output. [batch, output_nodes]
          - c: the lstm cell state. [batch, hidden_nodes]
      input_batch: A NestedMap containing:
        - src.src_inputs: tensor of shape [batch, len, dim].
        - src.paddings: tensor of shape [batch, len].
        - label: int tensor of shape [batch]. The inputs has ground truth labels
          required for the loss calculation.

    Returns:
      A tuple of two elements:
        - A NestedMap containing str keys and (metric, weight) values, where
          one of the keys is expected to be 'loss'.
        - A NestedMap containing arbitrary tensors describing something about
          each training example, where the first dimension of each tensor is
          the batch index.
    """
    # Inspect the batch to check duplicates before any transformation.
    self.batch_without_duplicate_speakers = self.BatchWithoutDuplicateSpeakers(
        input_batch)
    score = self.ComputeSimilarity(predictions.logits, theta)
    score, input_batch = self.TransformScoreAndLabel(theta, score, input_batch)
    metrics = self.CalculateLoss(theta, score, input_batch)
    return metrics, py_utils.NestedMap()


class GEnd2EndExtendedSetSoftmaxLayer(GEnd2EndSoftmaxLayer):
  """The generalized end-to-end extended-set softmax loss layer.

  This expands upon the standard row based softmax with localised cross-entropy.
  It provides improved score normalization (for speaker accept/reject
  thresholding purposes and better overall performance) by including non-target
  scores from other test segments. This encourages target scores to not only be
  large relative to the scores of the corresponding non-target models but also
  large relative to the scores of other test segment and model pairings. This
  code is currently designed for the standard text-independent data layout.

  To learn more about this loss, refer to this paper:
  https://arxiv.org/abs/2104.01989
  """

  @classmethod
  def Params(cls):
    p = super(GEnd2EndExtendedSetSoftmaxLayer, cls).Params()
    p.name = 'gend2end_extended_set_softmax_layer'
    return p

  def TransformScoreAndLabel(self, theta, score, input_batch):
    """Transformation with additional extended-set transformation.

    Args:
      theta: A NestedMap containing layer weights.
      score: tensor of shape [batch, num_spks_per_batch] with scores of each
        example to each speaker.
      input_batch: A NestedMap containing at least:
        - label: int tensor of shape
          [batch]. The inputs has ground truth labels required for the loss
          calculation.

    Returns:
      A tuple of two elements:
        - A transformed score tensor of shape [batch, num_proxy_classes]
        - A NestedMap input_batch containing the transformed label tensor of
          shape [batch, num_proxy_classes]
    """
    # Linear transformation and masking if specified.
    score, input_batch = super().TransformScoreAndLabel(theta, score,
                                                        input_batch)

    # Rework the scores
    refactored_scores, refactored_labels = self._RefactorExtendedSetInput(score)
    # New label tensor containing all zeros to represent that the zeroth indexed
    # column (1st column) in refactored_scores contains the target scores.
    input_batch.label = refactored_labels
    return refactored_scores, input_batch

  def _TransformStripedToStackedMatrix(self, striped_score_matrix):
    """Transforms a partially striped matrix to a stacked matrix.

    Given a striped score matrix (the striping pattern relates to the
    targets in the score matrix) it produces a stacked 'diagonal' matrix by
    swapping the rows of the input matrix.

    Args:
      striped_score_matrix: A 2-D tensor of size ([num_segments, num_models])
        containing the speaker scores.

    Returns:
      stacked_diagonal matrix: A 2-D tensor of stacked square matrices where
        each matrix consists of a diagonal of target scores and a
        non-diagonal of non-target scores.

    Explanation:
      For example, on the left is a striped score matrix with the target
        scores identified using round brackets (). On the right is the
        corresponding output from the function

          Row                     From Row
           1 [(11)  12   13 ]        1 [(11)  12   13 ]
           2 [(21)  22   23 ]        3 [ 31  (32)  33 ]
           3 [ 31  (32)  33 ]   -->  5 [ 51   52  (53)]
           4 [ 41  (42)  43 ]        2 [(21)  22   23 ]
           5 [ 51   52  (53)]        4 [ 41  (42)  43 ]
           6 [ 61   62  (63)]        6 [ 61   62  (63)]
    """

    # Determine the number of test segments and the number of models
    num_segs, num_models = [
        int(i) for i in striped_score_matrix.shape.as_list()
    ]

    # Calculate the number of test segments per speaker model
    # (verify that num_segs is a multiple of num_models)
    if num_segs % num_models != 0:
      raise ValueError(
          'Note that num_segs (%i) is not a multiple of num_models (%i).' %
          (num_segs, num_models))

    segs_per_speaker = num_segs // num_models

    # Specify the new row ordering
    row_ordering = tf.reshape(
        tf.transpose(
            tf.reshape(
                tf.range(0, num_segs, dtype=tf.int32),
                [segs_per_speaker, num_models])), [num_segs])

    # Apply the row ordering to the matrix
    stacked_diagonal_target_score_matrix = tf.scatter_nd(
        indices=tf.expand_dims(input=row_ordering, axis=1),
        updates=striped_score_matrix,
        shape=[num_segs, num_models])

    return stacked_diagonal_target_score_matrix

  def _ExtractTargetScoresFromStackedMatrix(self,
                                            stacked_diagonal_score_matrix):
    """Given a stacked 'diagonal' matrix, extracts corresponding target scores.

    Args:
      stacked_diagonal_score_matrix: A 2-D tensor of stacked square matrices
        where each matrix consists of a diagonal of target scores and a
        non-diagonal of non-target scores.

    Returns:
      tar_scores: A 1-D tensor of the target scores

    Explanation:
      For example, on the left is a stacked matrix and on the right are the
      corresponding target scores.

        Row
         1 [(11)  12   13 ]
         2 [ 31  (32)  33 ]
         3 [ 51   52  (53)]    -->   [(11) (32) (53) (21) (42) (63)]
         4 [(21)  22   23 ]
         5 [ 41  (42)  43 ]
         6 [ 61   62  (63)]
    """

    # Determine the number of test segments and the number of models
    num_segs, num_models = [
        int(i) for i in stacked_diagonal_score_matrix.shape.as_list()
    ]

    # Calculate the number of test segments per speaker model
    # (verify that num_segs is a multiple of num_models)
    if num_segs % num_models != 0:
      raise ValueError(
          'Note that num_segs (%i) is not a multiple of num_models (%i).' %
          (num_segs, num_models))

    segs_per_speaker = num_segs // num_models

    # Extract the target speaker scores
    tar_mask = tf.tile(tf.eye(num_models, dtype=tf.bool), [segs_per_speaker, 1])
    tar_scores = tf.boolean_mask(stacked_diagonal_score_matrix, tar_mask)

    return tar_scores

  def _ExtractNonTargetScoresForEachStackedMatrix(
      self, stacked_diagonal_score_matrix):
    """Extracts non-target scores from a stacked 'diagonal' matrix.

    For each stacked 'diagonal' matrix this function will produce a row of
    non-target scores. If there are N stacked matrices, there will be N rows.
    If the number of columns in the matrix is C, then the number of non-target
    scores for each row will be Cx(C-1).

    Args:
      stacked_diagonal_score_matrix: A 2-D tensor of stacked square matrices
        where each matrix consists of a diagonal of target scores and a
        non-diagonal of non-target scores.

    Returns:
      non_scores: A 1-D tensor of the non-target scores

    Explanation:
      For example, on the left is a stacked matrix and on the right are the
      corresponding non-target scores. The target scores are shown within
      brackets ().

        Row
         1 [(11)  12   13 ]
         2 [ 31  (32)  33 ]
         3 [ 51   52  (53)]    -->   [ 12  13  31  33  51  52]
         4 [(21)  22   23 ]          [ 22  23  41  43  61  62]
         5 [ 41  (42)  43 ]
         6 [ 61   62  (63)]
    """

    # Determine the number of test segments and the number of models
    num_segs, num_models = [
        int(i) for i in stacked_diagonal_score_matrix.shape.as_list()
    ]

    # Calculate the number of test segments per speaker model
    # (verify that num_segs is a multiple of num_models)
    if num_segs % num_models != 0:
      raise ValueError(
          'Note that num_segs (%i) is not a multiple of num_models (%i).' %
          (num_segs, num_models))

    segs_per_speaker = num_segs // num_models

    # Extract the non-target speaker scores
    non_mask = tf.tile(
        tf.math.logical_not(tf.eye(num_models, dtype=tf.bool)),
        [segs_per_speaker, 1])
    non_scores = tf.boolean_mask(stacked_diagonal_score_matrix, non_mask)

    # From each row of the output matrix, collect the non-target scores from
    # the same stacked 'diagonal' matrix.
    grouped_non_scores = tf.reshape(
        non_scores, [segs_per_speaker, num_models * (num_models - 1)])

    return grouped_non_scores

  def _RepeatRows(self, in_matrix, repeat_n_times):
    """Repeats each row of a matrix multiple times to produce a new matrix.

    Given a matrix this function will return a new matrix where each row
    becomes a repeated version of each row. For a PxQ input matrix, the
    output matrix will now have P times N rows and Q columns. An example
    follows below.

    Args:
      in_matrix: A PxQ matrix.
      repeat_n_times: How many times to repeat each row within the matrix.

    Returns:
      repeated_rows_matrix: A matrix calculated from in_matrix by repeating its
      rows.

    Explanation:
      For example, on the left is the input matrix (in_matrix), and
      repeat_n_times is set to 2. This produces the resulting matrix on the
      right.

        Row                        Row
         1 [(11)  12   13 ]         1 [(11)  12   13 ]
         2 [ 31  (32)  33 ]         2 [(11)  12   13 ]
         3 [ 51   52  (53)]    -->  3 [ 31  (32)  33 ]
                                    4 [ 31  (32)  33 ]
                                    5 [ 51   52  (53)]
                                    6 [ 51   52  (53)]
    """

    # Determine the number of matrix rows and columns
    num_rows, num_cols = [int(i) for i in in_matrix.shape.as_list()]

    # Horizontally stack the matrix multiple times
    stacked_matrix = tf.tile(
        input=in_matrix, multiples=tf.constant([1, repeat_n_times]))

    # Reshape the matrix to produce the final repeated rows
    repeated_rows_matrix = tf.reshape(stacked_matrix,
                                      [repeat_n_times * num_rows, num_cols])

    return repeated_rows_matrix

  def _RefactorExtendedSetInput(self, score):
    """Refactors the input score matrix into extended-set softmax form.

    Transform the scores such that the first column has the same-speaker
    scores and other columns have the (extended-set) different-speaker
    scores.

    Args:
      score: A 2-D tensor of size ([num_segments, num_models]) containing the
        speaker scores.

    Returns:
      final_scores: A 2-D tensor of scores with the first column containing
        target scores and the other columns containing the non-target scores.
        Size: ([num_segments, num_models*(num_models-1)+1])
      final_labels: A same-speaker label vector of zeros indicating that the
        same-speaker score is in the first column of final_scores.
    """

    # Determine the number of test segments and the number of models
    num_segs, num_models = [int(i) for i in score.shape.as_list()]

    # Convert the striped score matrix into a stacked "diagonal" matrix
    # Input shape: score = [num_segs x num_models]
    # Output shape: stacked_matrix = [num_segs x num_models]
    stacked_matrix = self._TransformStripedToStackedMatrix(score)

    # Extract the target scores from a stacked "diagonal" matrix
    # Input shape: stacked_matrix = [num_segs x num_models]
    # Output shape: tar_scores = [1 x num_segs]
    tar_scores = self._ExtractTargetScoresFromStackedMatrix(stacked_matrix)

    # Extract the non-target scores from a stacked "diagonal" matrix.
    # The difference this time is that the non-target scores for each
    # stacked square matrix are collected as a single row in the output
    # matrix.
    # Input shape: stacked_matrix = [num_segs x num_models]
    # Output shape: grouped_non_scores =
    #   [(num_segs/num_models) x (num_models*(num_models-1))]
    grouped_non_scores = self._ExtractNonTargetScoresForEachStackedMatrix(
        stacked_matrix)

    # Repeat the contents of each row multiple times to correspond to
    # the target trials in each stacked matrix
    # Input shape: grouped_non_scores =
    #   [(num_segs/num_models) x (num_models*(num_models-1))]
    # Output shape: repeated_non_scores =
    #   [num_segs x (num_models*(num_models-1))]
    repeated_non_scores = self._RepeatRows(grouped_non_scores, num_models)

    # Place the target scores at the beginning and the non-targets second
    # Input shapes: #1 [num_seg x 1]
    #               #2 [num_segs x (num_models*(num_models-1))]
    # Output shape: final_scores [num_seg x (num_models*(num_models-1)+1)]
    final_scores = tf.concat(
        values=[tf.expand_dims(tar_scores, 1), repeated_non_scores], axis=1)

    # Target trials are in column 0, so create a label key with a zeros
    final_labels = tf.zeros([num_segs], dtype=tf.int32)

    return final_scores, final_labels



"""Tests for loss_layers."""

from lingvo import compat as tf
from lingvo.core import py_utils
from lingvo.core import test_utils

import numpy as np
from sidlingvo import loss_layers

FLAGS = tf.flags.FLAGS
_RANDOM_SEED = 1314


class GEnd2EndSoftmaxLayerTest(test_utils.TestCase):

  BATCH_SIZE = 6
  NUM_SPKS_PER_BATCH = 3
  NUM_UTTS_PER_SPK = 2

  def setUp(self):
    super().setUp()
    tf.reset_default_graph()

    # test data
    logits = tf.convert_to_tensor(
        np.array([[1.0, 1.0], [2.0, 1.0], [3.0, 1.0], [1.0, 2.0], [2.0, 2.0],
                  [3.0, 2.0]]),
        dtype=tf.float32)
    labels = tf.convert_to_tensor(np.array([0, 0, 1, 1, 2, 2]), dtype=tf.int32)
    self.predictions = py_utils.NestedMap(logits=logits)
    self.input_batch = py_utils.NestedMap(label=labels)
    # scores for testing the duplicate speaker masking
    self.scores = tf.convert_to_tensor(
        np.array([[1.0, 0.1, 0.8], [2.0, 1.0, 1.8], [0.1, 1.0, 0.2],
                  [0.2, 1.1, 0.1], [2.0, 0.2, 2.0], [3.0, 0.3, 2.0]]),
        dtype=tf.float32)

  def _PrepareLossLayerParam(self):
    p = loss_layers.GEnd2EndSoftmaxLayer.Params()
    p.name = 'test'
    p.random_seed = 123

    p.batch_size = self.BATCH_SIZE
    p.num_spks_per_batch = self.NUM_SPKS_PER_BATCH
    p.num_utts_per_spk = self.NUM_UTTS_PER_SPK
    p.softmax.input_dim = self.NUM_SPKS_PER_BATCH
    p.softmax.num_classes = self.NUM_SPKS_PER_BATCH

    return p

  def testMaskDuplicateSpeakerScoresOnDuplicates(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = True
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      # Overwrite the speaker labels to have duplicate speakers
      labels = tf.convert_to_tensor(
          np.array([0, 0, 1, 1, 0, 0]), dtype=tf.int32)
      self.input_batch = py_utils.NestedMap(label=labels)
      masked_scores, masked_input_batch = sess.run(
          test_layer.MaskDuplicateSpeakerScores(self.scores, self.input_batch))
      expected_scores = np.array(
          [[1.0, 0.1, -99.2], [2.0, 1.0, -98.2], [0.1, 1.0, 0.2],
           [0.2, 1.1, 0.1], [-98.0, 0.2, 2.0], [-97.0, 0.3, 2.0]], np.float32)
      expected_labels = np.array([0, 0, 1, 1, 2, 2], np.int32)
      self.assertAllEqual(masked_scores, expected_scores)
      self.assertAllEqual(masked_input_batch.label, expected_labels)

  def testMaskDuplicateSpeakerScoresOnUniques(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = True
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      (masked_scores,
       masked_input_batch), expected_scores, expected_labels = sess.run([
           test_layer.MaskDuplicateSpeakerScores(self.scores, self.input_batch),
           self.scores, self.input_batch.label
       ])
      # Expect no-op
      self.assertAllEqual(masked_scores, expected_scores)
      self.assertAllEqual(masked_input_batch.label, expected_labels)

  def testFPropWithNoSplit(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = False
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      metrics, _ = sess.run(
          test_layer.FPropDefaultTheta(self.predictions, self.input_batch))

    self.assertAlmostEqual(1.0893, metrics.loss[0], places=4)

  def testFPropWithSplit(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = True
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      metrics, _ = sess.run(
          test_layer.FPropDefaultTheta(self.predictions, self.input_batch))

    # Expect higher average loss than the no split scenario.
    self.assertAlmostEqual(1.8346, metrics.loss[0], places=4)

  def testFPropWithDuplicateSpeakers(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = True
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      # Overwrite the speaker labels to have duplicate speakers
      labels = tf.convert_to_tensor(
          np.array([0, 0, 1, 1, 0, 0]), dtype=tf.int32)
      self.input_batch = py_utils.NestedMap(label=labels)
      metrics, _ = sess.run(
          test_layer.FPropDefaultTheta(self.predictions, self.input_batch))

    # Expect the loss weight to be 0, no matter what the loss is.
    self.assertAllEqual(0, metrics.loss[1])

  def testFPropWithDuplicateSpeakersMask(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = True
      p.mask_dup_spk_scores = True
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      # Overwrite the speaker labels to have duplicate speakers
      labels = tf.convert_to_tensor(
          np.array([0, 0, 1, 1, 0, 0]), dtype=tf.int32)
      self.input_batch = py_utils.NestedMap(label=labels)
      metrics, _ = sess.run(
          test_layer.FPropDefaultTheta(self.predictions, self.input_batch))

    # Expect lower average loss than the no masking scenario.
    self.assertAlmostEqual(1.5610, metrics.loss[0], places=4)

    # Expect the loss weight to be non-zero under masking.
    self.assertAllEqual(6, metrics.loss[1])


class GEnd2EndExtendedSetSoftmaxLayerTest(GEnd2EndSoftmaxLayerTest):

  def _PrepareLossLayerParam(self):
    p = loss_layers.GEnd2EndExtendedSetSoftmaxLayer.Params()
    p.name = 'test'
    p.random_seed = 123

    p.batch_size = self.BATCH_SIZE
    p.num_spks_per_batch = self.NUM_SPKS_PER_BATCH
    p.num_utts_per_spk = self.NUM_UTTS_PER_SPK

    num_proxy_classes = self.NUM_SPKS_PER_BATCH * (self.NUM_SPKS_PER_BATCH -
                                                   1) + 1
    p.softmax.input_dim = num_proxy_classes
    p.softmax.num_classes = num_proxy_classes

    return p

  def testFPropWithNoSplit(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = False
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      metrics, _ = sess.run(
          test_layer.FPropDefaultTheta(self.predictions, self.input_batch))

    self.assertAlmostEqual(1.9803, metrics.loss[0], places=4)

  def testFPropWithAttentionScoring(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = True
      p.select_embedding_comparison_type = (
          loss_layers.EmbeddingComparisonType.ATTENTIVE_SCORING)
      p.attentive_scoring.num_keys = 1
      p.attentive_scoring.key_dim = 1
      p.attentive_scoring.value_dim = 1
      print(self.input_batch)

      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      metrics, _ = sess.run(
          test_layer.FPropDefaultTheta(self.predictions, self.input_batch))

    self.assertAlmostEqual(1.94591, metrics.loss[0], places=4)

  def testFPropWithSplit(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = True
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      metrics, _ = sess.run(
          test_layer.FPropDefaultTheta(self.predictions, self.input_batch))

    # Expect higher average loss than the no split scenario.
    self.assertAlmostEqual(2.7612, metrics.loss[0], places=4)

  def testFPropWithDuplicateSpeakersMask(self):
    with self.session() as sess:
      tf.random.set_seed(_RANDOM_SEED)
      p = self._PrepareLossLayerParam()
      p.split_batch = True
      p.mask_dup_spk_scores = True
      test_layer = p.Instantiate()
      tf.global_variables_initializer().run()
      # Overwrite the speaker labels to have duplicate speakers
      labels = tf.convert_to_tensor(
          np.array([0, 0, 1, 1, 0, 0]), dtype=tf.int32)
      self.input_batch = py_utils.NestedMap(label=labels)
      metrics, _ = sess.run(
          test_layer.FPropDefaultTheta(self.predictions, self.input_batch))

    # Expect lower average loss than the no masking scenario.
    self.assertAlmostEqual(2.4000, metrics.loss[0], places=4)

    # Expect the loss weight to be non-zero under masking.
    self.assertAllEqual(6, metrics.loss[1])


if __name__ == '__main__':
  tf.test.main()
